/*
   Version using SSE 'ATLAS' for matrix multiplication.
   Makefile, for example:
------------------------------------------------------------------------
ALIB = /home/zibi/install/ATLAS/lib/
AINC = /home/zibi/install/ATLAS/include/
 
mpeaks: mpeaks.c
        gcc  -Wall -O3 $< -I $(AINC) -L$(ALIB) \
        -lcblas -latlas -lm -ffast-math -o $@
------------------------------------------------------------------------
 
     where /home/zibi/install/ATLAS/ is path to 'ATLAS',
     see ATLAS homepage:
     http://math-atlas.sourceforge.net	   
*/

#include<stdio.h>
#include<malloc.h>
#include<stdlib.h>
#include<math.h>
#include<sys/stat.h>
#include<string.h>
//#include"string.h"
#include "cblas.h"

#if !defined(MaxNofChars)
#define MaxNofChars 456
#endif

#define N_STARS 500

static float startfr = 0.0, endfr = 30.0, stepfr = -1.0, delfr =
    0.2, centrfr = -1.0;
static float kappapeak = 1.3, maxdt = 0.00005, kappaclean = 9.0;
static int trfidx = 0, cleanidx = 0, appendidx = 0, minobs = 30;

static char *logname = "mpeaks.log";

char getln(file, line)
FILE *file ; char *line ;
 {
 int i=0 ;
 for( ; (*line=getc(file))!='\n' && !feof(file) ; line++, i++)
   if(i==MaxNofChars)
     { printf("  Wrn: getln: line to big! Break.\n"); break ; }
 *line=0 ;
 if(feof(file)) return(0);
 return(1);
 }

void getext(file, ext)
char *file ; char *ext ;
 {
 int i ;
 for(i=0 ; *file ; file++, i++);
 for(file--, i-- ; i>=0 ; file--, i--)
   if(*file=='.') break ;
 *ext=0 ;
 if(*file=='.')
   {
   for(i++ ; (*ext++=*++file)!=0 ; i++);
   for(file--, i-- ; i>=0 ; file--, i--)
     if(*file=='.') { *file=0 ; break ; }
   }
 }

void strcopy(str, copy)
char *str ; char *copy ;
 {
 while((*str++=*copy++)!=0);
 }

void stradd(str, add)
char *str ; char *add ;
 {
 while(*str) str++ ;
 while((*str++=*add++)!=0);
 }

char strcomp(str1, str2)
char *str1 ; char *str2 ;
 {
 if(strlen(str1)!=strlen(str2)) return(0);
 for( ; *str1==*str2 ; str1++, str2++)
   if(!*str1) return(1);
 return(0);
 }

char whitech(c)
char c ;
 { 
 return((c==' '||c=='\t')? 1 : 0);
 }

void whitestrip(str)
char *str ;
 {
 int istr, itemp ;
 char *temp=(char *)malloc(strlen(str)+1) ;
 if(!temp) return ;
 for(istr=0 ; str[istr] && whitech(str[istr]) ; istr++);
 if(!str[istr]) { *str=0 ; return ; }
 for(itemp=0 ; str[istr] ; istr++, itemp++) temp[itemp]=str[istr] ;
 for(istr--, itemp-- ; whitech(str[istr]) && istr>=0 ; istr--, itemp--);
 temp[++itemp]=0 ;
 for(itemp=0 ; temp[itemp] ; itemp++) str[itemp]=temp[itemp] ;
 str[itemp]=0 ;
 free(temp);
 }

char strpcomp(str, text) char *str, *text ;
 {
 if(strlen(str)>strlen(text)) return(0) ;
 while(*str)
   if(*str++!=*text++) return(0);
 return(1);
 }

int searchstr(text, str) char *text, *str ;
 {
 int ifind=0, lenofstr=strlen(str) ;
 if(lenofstr>strlen(text)) return(0);
 while(*text) 
   if(strpcomp(str,text)) { text+=lenofstr ; ifind++ ; }
   else text++ ;
 return(ifind);
 }

void usage(void)
{
    printf
	("\n MPEAKS for Linux\n by W. Hebisch, Z. Kolaczkowski, G. Kopacki.\n");
    printf(" Version of Mar 28, 2010.\n\n");
    printf(" Usage: mpeaks [<options>] <times> <datacat>\n");
    printf(" <options>:\n");
    printf("   -f=<first_fr>,<last_fr>,<fr_step>   "
	   "(default: %.0f,%.0f,computed)\n", startfr, endfr);
    printf("   -f=<first_fr>,<last_fr>   (<fr_step> is computed)\n");
    printf("   -fc=<centr_fr>,<delta_fr>,<fr_step>\n");
    printf("   -fc=<centr_fr>,<delta_fr>   (<fr_step> is computed)\n");
    printf
	("   -fc=<centr_fr>   (<delta_fr>=%.1f, <fr_step> is computed)\n",
	 delfr);
    printf("   -append   (data are appended to *.max and *.trf files)\n");
    printf("   -trf   (spectrum is written into *.trf file)\n");
    printf("   -dt=<max_tim_diff>   (in days)\n");
    printf("   -minobs=<min_no_of_obs>\n");
    printf("   -clean   (light-curve cleaning)\n");
    printf("   -k=<kappa_clean>\n");
    printf(" Input file format: <time> <signal> [omitted fields]\n");
    printf(" Most prominent peaks are written into *.max file\n");
    printf(" Some important parameters are written into log-file \"%s\"\n\n",
	   logname);
}

void sortpeaks(fr, amp, n)
register double *fr, *amp;
int n;
{
    double temp;
    register int gap, i, j;
    for (gap = n / 2; gap > 0; gap /= 2)
	for (i = gap; i < n; i++)
	    for (j = i - gap; j >= 0 && amp[j] < amp[j + gap]; j -= gap) {
		temp = amp[j];
		amp[j] = amp[j + gap];
		amp[j + gap] = temp;
		temp = fr[j];
		fr[j] = fr[j + gap];
		fr[j + gap] = temp;
	    }
}

struct SINCOS {
    float s, c;
};

long memusage = 0;

void *my_malloc(long s)
{
    void *res = malloc(s);
    if (!res) {
	fprintf(stderr, "Out of memory for %ld bytes\n", s);
	exit(1);
    }
    memusage += s;
    return res;
}


typedef struct freq_data {
    /* Work areas */
    double *sintab, *costab;
    /* Start frequency sine/cosine table */
    double *sintab0, *costab0;
    /* One step frequency sine/cosine table */
    double *sintab1, *costab1;
    /* Number of time points */
    int Ntime;
    /* Number of frequencies to calculate */
    int Namp;
    /* Work area for sine/cosine table */
    float **matrix;
    /* Table of time points */
    double *time;
    /* Work area for a batch of data points */
    float *obs_tab;
    /* Work area  */
    float *amp_tab;

    double avramp_tab[N_STARS];
    int Nobs[N_STARS];
    float *cs_pro;
    /* Work area */
    char *idx;
    char *name_tab[N_STARS];
} freq_data;

int group_data_processing(FILE * cat, FILE * log, freq_data * f_data);

FILE *file;
char name[256], line[256];


int main(narg, arg)
int narg;
char **arg;
{
    FILE *cat, *log;
    double mint, maxt;
    int Ntime=0;
    int Namp;

    int i;
    freq_data *f_data;

    if (narg < 3) {
	usage();
	return (0);
    }

    for (i = 0; i < narg; i++) {
	if (**++arg != '-')
	    break;
	if (sscanf(*arg, "-f=%f,%f,%f", &startfr, &endfr, &stepfr) == 3) {
	    continue;
	}
	if (sscanf(*arg, "-f=%f,%f", &startfr, &endfr) == 2) {
	    continue;
	}
	if (sscanf(*arg, "-fc=%f,%f,%f", &centrfr, &delfr, &stepfr) == 3) {
	    continue;
	}
	if (sscanf(*arg, "-fc=%f,%f", &centrfr, &delfr) == 2) {
	    continue;
	}
	if (sscanf(*arg, "-fc=%f", &centrfr) == 1) {
	    continue;
	}
	if (sscanf(*arg, "-k=%f", &kappaclean) == 1) {
	    continue;
	}
	if (sscanf(*arg, "-minobs=%d", &minobs) == 1) {
	    continue;
	}
	if (sscanf(*arg, "-dt=%f", &maxdt) == 1) {
	    continue;
	}
	if (strcomp(*arg, "-trf")) {
	    trfidx++;
	    continue;
	}
	if (strcomp(*arg, "-clean")) {
	    cleanidx++;
	    continue;
	}
	if (strcomp(*arg, "-append")) {
	    appendidx++;
	    continue;
	}
    }
    narg -= i;

    if (narg != 3) {
	usage();
	return (0);
    }

    f_data = my_malloc(sizeof(*f_data));
    log = fopen(logname, (appendidx) ? "at" : "wt");
    if (log) {
    
    	fprintf(log,"%% StartFr = %.3f   EndFr = %.3f   StepFr = %.6f   NTime = %d\n",
	   startfr, endfr, stepfr, Ntime);		
	fprintf(log,"%%       Name         Nobs         Freq          Period       Ampl      S/N\n"); 
    }

    strcopy(name, *arg);
    if (!(file = fopen(name, "rt"))) {
	printf("  Error: Data file \"%s\" not open for reading!\n", name);
	return (0);
    }

    for (i = 0; !feof(file);) {
	getln(file, line);
	if (*line != '%' && *line)
	    i++;
    }
    fseek(file, 0, 0);

    if (!(f_data->time = (double *) calloc(i, sizeof(double)))) {
	printf("  Error: Not enough memory for %d items\n", i);
	return (0);
    }
    if (!(f_data->idx = (char *) calloc(i, sizeof(char)))) {
	printf("  Error: Not enough memory for %d items\n", i);
	return (0);
    }
    memusage += i * (2 * sizeof(double) + sizeof(char));

    for (i = 0; !feof(file);) {
	do
	    getln(file, line);
	while (*line == '%' || searchstr(line, "nan"));
	if (sscanf(line, "%lf", &(f_data->time[i])) != 1)
	    break;
	i++;
    }
    f_data->Ntime = Ntime = i;
    fclose(file);

    if (stepfr <= 0.0) {
	mint = maxt = f_data->time[0];
	for (i = 1; i < Ntime; i++) {
	    double atim = f_data->time[i];
	    if (atim > maxt)
		maxt = atim;
	    if (atim < mint)
		mint = atim;
	}
	stepfr = 0.2 / (maxt - mint);
    }

    if (centrfr > 0.0) {
	if (delfr < 0.001)
	    delfr = 0.1;
	if ((startfr = centrfr - delfr) < 0.0)
	    startfr = 0.0;
	endfr = centrfr + delfr;
    }

    f_data->Namp = Namp = (int) ((endfr - startfr) / stepfr);
    
    printf(">> StartFr = %.2f   EndFr = %.2f   StepFr = %.6f   NTime = %d <<\n",
	   startfr, endfr, stepfr, Ntime);

    f_data->matrix = my_malloc(2 * 100 * sizeof(float *));
    {
	float *ptab = my_malloc(2 * 100 * Ntime * sizeof(float));
	int i;
	for (i = 0; i < 2 * 100; i++) {
	    f_data->matrix[i] = ptab;
	    ptab += Ntime;
	}
    }

    /* Initializing sin and cos tables */
    f_data->sintab = my_malloc(sizeof(*f_data->sintab) * Ntime);
    f_data->costab = my_malloc(sizeof(*f_data->costab) * Ntime);
    f_data->sintab0 = my_malloc(sizeof(*f_data->sintab0) * Ntime);
    f_data->costab0 = my_malloc(sizeof(*f_data->costab0) * Ntime);
    f_data->sintab1 = my_malloc(sizeof(*f_data->sintab1) * Ntime);
    f_data->costab1 = my_malloc(sizeof(*f_data->costab1) * Ntime);
    for (i = 0; i < Ntime; i++) {
	double startfreq = 2 * M_PI * startfr;
	double stepfreq = 2 * M_PI * stepfr;
	f_data->sintab0[i] = sin(startfreq * f_data->time[i]);
	f_data->costab0[i] = cos(startfreq * f_data->time[i]);
	f_data->sintab1[i] = sin(stepfreq * f_data->time[i]);
	f_data->costab1[i] = cos(stepfreq * f_data->time[i]);
    }

    if (!(cat = fopen(*++arg, "rt"))) {
	printf("  Error: Data file \"%s\" not open for reading!\n", *arg);
	return (0);
    }

    {
	f_data->obs_tab = my_malloc(N_STARS * Ntime *
				    sizeof(*f_data->obs_tab));
	f_data->amp_tab = my_malloc(N_STARS * Namp *
				    sizeof(*f_data->amp_tab));
	f_data->cs_pro = my_malloc(2 * 100 * N_STARS *
				   sizeof(*f_data->cs_pro));

	while (!group_data_processing(cat, log, f_data));
	printf("\n");

	fclose(cat);
	if (log)
	    fclose(log);


	free(f_data->time);
	free(f_data->idx);

	return (0);
    }
}

int group_data_processing(FILE * cat, FILE * log, freq_data * f_data)
{
    int i, j;
    int star_no = 0;
    int Ntime = f_data->Ntime;
    int Namp = f_data->Namp;
    float *obs;
    char *idx = f_data->idx;
    char ext[256], catline[256];
    int Npeak;
    float *amp;
    double *frequency, *peak;
    double t, m;
    double sum, freq, avrobs, avramp;
    FILE * file;
    
    printf("."); fflush(stdout);
    while (!feof(cat) && star_no < N_STARS) {
	int Nobs = 0;
	obs = &(f_data->obs_tab[star_no * Ntime]);
	do
	    getln(cat, catline);
	while (*catline == '%');
	whitestrip(catline);
	if (!*catline)
	    break;

	for (i = 0; i < Ntime; i++)
	    idx[i] = 0;

	strcpy(name, catline);
	f_data->name_tab[star_no] = malloc(strlen(name) + 1);
	strcpy(f_data->name_tab[star_no], name);
	if (!(file = fopen(name, "rt"))) {
	    printf("  Error: Data file \"%s\" not open for reading!\n",
		   name);
	    free(f_data->name_tab[star_no]);
	    continue;
	}

	avrobs = Nobs = 0;

//  Reading data...
	while (!feof(file)) {
	    do
		getln(file, line);
	    while (*line == '%' || searchstr(line, "nan"));
	    if (sscanf(line, "%lf%lf", &t, &m) != 2)
		break;
	    for (i = 0; i < Ntime; i++)
		if (fabs(f_data->time[i] - t) < maxdt) {
		    obs[i] = m;
		    idx[i]++;
		    Nobs++;
		    avrobs += m;
		    break;
		}
	}
	fclose(file);
	if (Nobs < minobs) {
	    free(f_data->name_tab[star_no]);
	    continue;
	}
	
	f_data->Nobs[star_no] = Nobs;
	avrobs /= (double) Nobs;

//   for(i=0 ; i<Ntime ; i++) 
//   printf(" %.5f %.3f %1d\n", time[i], obs[i], idx[i]);

	if (cleanidx) {
//     printf(" Cleaning...");
	    for (sum = i = 0; i < Ntime; i++)
		if (idx[i])
		    sum += (obs[i] - avrobs) * (obs[i] - avrobs);
	    sum = sqrt(sum / Nobs);
	    for (j = i = 0; i < Ntime; i++)
		if (idx[i] && fabs(obs[i] - avrobs) > kappaclean * sum) {
		    idx[i] = 0;
		    j++;
		}
//     printf("  %d points discarded.\n", j);
	    for (avrobs = Nobs = i = 0; i < Ntime; i++)
		if (idx[i]) {
		    avrobs += obs[i];
		    Nobs++;
		}
	    avrobs /= (double) Nobs;
	}

	for (i = 0; i < Ntime; i++)
	    if (idx[i]) {
		obs[i] -= avrobs;
	    } else {
		obs[i] = 0.0;
	    }
	star_no++;
    }
    if (!star_no) {
	return 1;
    }
    
    {
	double *costab = f_data->costab;
	double *sintab = f_data->sintab;
	double *costab0 = f_data->costab0;
	double *sintab0 = f_data->sintab0;

	for (j = 0; j < Ntime; j++) {
	    costab[j] = costab0[j];
	    sintab[j] = sintab0[j];
	}
    }

//   printf(" ObsAvr = %.4f   NofObs = %d\n", avrobs, Nobs);
    {
	int star_ind;
	float *cs_pro = f_data->cs_pro;
	for (star_ind = 0; star_ind < star_no; star_ind++) {
	    f_data->avramp_tab[star_ind] = 0;
	}
	for (j = 0; j < Namp; j += 100) {
	    int jj_step = (Namp - j > 100) ? 100 : Namp - j;
	    int j1;
	    for (j1 = 0; j1 < 100; j1++) {
		float *mpc = f_data->matrix[2 * j1];
		float *mps = f_data->matrix[2 * j1 + 1];
		double *costab = f_data->costab;
		double *sintab = f_data->sintab;
		double *costab1 = f_data->costab1;
		double *sintab1 = f_data->sintab1;

		int i;
		freq = 2 * M_PI * (startfr + j1 * stepfr);
		for (i = 0; i < Ntime; i++) {
		    double cosv = costab[i];
		    double sinv = sintab[i];
		    costab[i] = cosv * costab1[i] - sinv * sintab1[i];
		    sintab[i] = sinv * costab1[i] + cosv * sintab1[i];
		    *mpc = cosv;
		    *mps = sinv;
		    mps++;
		    mpc++;
		}
	    }

	    /* Compute sin and cosine sums using matrix multiplication */
	    cblas_sgemm(CblasRowMajor, CblasNoTrans, CblasTrans,
			star_no, 2 * jj_step, Ntime, 1.0,
			f_data->obs_tab, Ntime, f_data->matrix[0],
			Ntime, 0.0, cs_pro, 2 * 100);
	    for (star_ind = 0; star_ind < star_no; star_ind++) {
		int jj;
		int Nobs = f_data->Nobs[star_ind];
		amp = &(f_data->amp_tab[star_ind * Namp + j]);
		avramp = 0;
		for (jj = 0; jj < jj_step; jj++) {
		    float cobs = cs_pro[star_ind * 200 + 2 * jj];
		    float sobs = cs_pro[star_ind * 200 + 2 * jj + 1];
		    avramp += *amp++ =
			2 * sqrt(cobs * cobs + sobs * sobs) / Nobs;
		}
		f_data->avramp_tab[star_ind] += avramp;
	    }
	}
    }

    avramp /= Namp;
    amp -= Namp;
    {
	int star_ind;
	for (star_ind = 0; star_ind < star_no; star_ind++) {
	    obs = &(f_data->obs_tab[star_ind * Ntime]);
	    amp = &(f_data->amp_tab[star_ind * Namp]);
	    avramp = f_data->avramp_tab[star_ind];
	    strcpy(name, f_data->name_tab[star_ind]);
	    free(f_data->name_tab[star_ind]);
	    Npeak = 0;
	    for (sum = i = 0; i < Namp; amp++, i++)
		if (*amp < 3.5 * avramp) {
		    sum += *amp;
		    Npeak++;
		}
	    avramp = sum / Npeak;
	    amp -= Namp;

	    Npeak = 0;
	    for (amp++, i = 1; i < Namp - 1; amp++, i++) {
		if (*amp < avramp * kappapeak || *amp < amp[-1]
		    || *amp < amp[+1])
		    continue;
		Npeak++;
	    }
	    amp++;
	    amp -= Namp;

	    if (Npeak < 2)
		continue;

	    if (!(frequency = (double *) calloc(Npeak, sizeof(double)))) {
		printf("  Error: Not enough memory for %d items\n", Npeak);
		return (0);
	    }
	    if (!(peak = (double *) calloc(Npeak, sizeof(double)))) {
		printf("  Error: Not enough memory for %d items\n", Npeak);
		return (0);
	    }

	    j = 0;
	    for (amp++, i = 1; i < Namp - 1; amp++, i++) {
		if (*amp < avramp * kappapeak || *amp < amp[-1]
		    || *amp < amp[+1])
		    continue;
		frequency[j] = startfr + i * stepfr;
		peak[j++] = *amp;
	    }
	    amp++;
	    amp -= Namp;

//   printf(" AmpAvr = %.5f   NofPeaks = %d\n", avramp, Npeak); 
//   fflush(stdout);

	    sortpeaks(frequency, peak, Npeak);

	    strcpy(catline, name);
	    getext(name, ext);
	    stradd(name, ".max");
	    file = fopen(name, (appendidx) ? "at" : "wt");
	    if (file) {
		fprintf(file, "%% Output of AmpPer program\n%%\n");
		fprintf(file,
			"%%  Most prominent peaks for star file : \"%s\" (%d points)\n",
			catline, f_data->Nobs[star_ind]);
		fprintf(file, "%%  Frequency range = ( %.2f , %.2f )\n",
			startfr, endfr);
		fprintf(file, "%%  Frequency step = %.5f\n", stepfr);
		fprintf(file, "%%  Noise estimate = %.5f\n%%\n", avramp);
		fprintf(file,
			"%%  No  Frequency       Period      Amplitude   S/N\n%%\n");
	    }
	    for (i = 0; i < Npeak; frequency++, peak++, i++) {
		if (file && i < 10)
		    fprintf(file, "  %3d %9.5f  %13.7f  %11.5f %8.3f\n",
			    i + 1, *frequency, 1 / (*frequency), *peak,
			    *peak / avramp);
	    }
	    frequency -= Npeak;
	    peak -= Npeak;
	    if (file)
		fclose(file);

	    if (log)
		    fprintf(log,
		           "%-20s %5d   %13.5f %13.5f %11.5f %8.3f\n",
			    catline, f_data->Nobs[star_ind], *frequency,
			    1/(*frequency), *peak, *peak/avramp);

		    
	    free(peak);
	    free(frequency);

	    if (trfidx) {
		strcopy(name, catline);
		getext(name, ext);
		stradd(name, ".trf");
		file = fopen(name, (appendidx) ? "at" : "wt");
		if (file) {
//       printf((appendidx)?" Appending spectrum...\n":" Writing spectrum...\n"); fflush(stdout);
		    fprintf(file, "%% Output of AmpPer program\n%%\n");
		    fprintf(file,
			    "%%  Amplitude spectrum for star file : \"%s\" (%d points)\n",
			    catline, f_data->Nobs[star_ind]);
		    fprintf(file,
			    "%%  Frequency range = ( %.2f , %.2f )\n",
			    startfr, endfr);
		    fprintf(file, "%%  Frequency step = %.5f\n", stepfr);
		    fprintf(file, "%%  Noise estimate = %.5f\n%%\n",
			    avramp);
		    fprintf(file,
			    "%%   Frequency      FourierAmplitude\n%%\n");
		    for (i = 0; i < Namp; i++)
			fprintf(file, "%.5f %.4f\n", startfr + i * stepfr,
				amp[i]);
		    fclose(file);
		}
	    }
	}
    }
    return N_STARS - star_no;
}
